import logging

from .base import BaseRule

logger = logging.getLogger(__name__)


class PercentageFailuresRule(BaseRule):

    def __init__(
        self,
        max_failures_percentage,
        failure_cache_key,
        min_accepted_requests,
        request_cache_key,
    ):
        super().__init__(
            failure_cache_key=failure_cache_key,
            request_cache_key=request_cache_key,
        )
        self.max_failures_percentage = max_failures_percentage
        self.min_accepted_requests = min_accepted_requests

    def _get_percentage_failures(self, total_failures, total_requests):
        if total_requests > 0:
            return (total_failures * 100) / total_requests
        return 0

    def should_open_circuit(self, total_failures, total_requests):
        percentage_failures = self._get_percentage_failures(
            total_failures=total_failures,
            total_requests=total_requests
        )
        return (
            total_requests > self.min_accepted_requests and
            percentage_failures >= self.max_failures_percentage
        )

    def log_increase_failures(self, total_failures, total_requests):
        percentage_failures = self._get_percentage_failures(
            total_failures=total_failures,
            total_requests=total_requests
        )
        logger.info(
            f'Increase failure for: {self.failure_cache_key} - '
            f'max failures {self.max_failures_percentage}% - '
            f'total failures {total_failures} - '
            f'min accepted requests {self.min_accepted_requests} - '
            f'total requests {total_requests} - '
            f'percentage failures {percentage_failures}%'
        )



import inspect
import logging

logger = logging.getLogger(__name__)


class CircuitBreakerBase:

    def __init__(
        self,
        rule,
        cache,
        failure_exception,
        failure_timeout=None,
        circuit_timeout=None,
        catch_exceptions=None,
    ):
        self.rule = rule
        self.cache = cache
        self.failure_timeout = failure_timeout
        self.circuit_timeout = circuit_timeout
        self.circuit_cache_key = 'circuit_{}'.format(rule.failure_cache_key)
        self.failure_exception = failure_exception
        self.catch_exceptions = catch_exceptions or (Exception,)

    def _is_catchable_exception(self, exception):
        return inspect.isclass(exception) and any(
            issubclass(exception, exception_class)
            for exception_class in self.catch_exceptions
        )

    def _notify_open_circuit(self):
        logger.critical(
            f'Open circuit for {self.rule.failure_cache_key} '
            f'{self.circuit_cache_key}'
        )

    def _notify_max_failures_exceeded(self):
        logger.info(f'Max failures exceeded by: {self.rule.failure_cache_key}')



import asyncio
from functools import wraps

from .base import CircuitBreakerBase


class CircuitBreaker(CircuitBreakerBase):

    async def is_circuit_open(self):
        return await self.cache.get(self.circuit_cache_key) or False

    async def get_total_failures(self):
        return await self.cache.get(self.rule.failure_cache_key) or 0

    async def get_total_requests(self):
        return await self.cache.get(self.rule.request_cache_key) or 0

    async def open_circuit(self):
        await self.cache.set(
            self.circuit_cache_key,
            1,
            self.circuit_timeout
        )

        # Delete the cache key to mitigate multiple sequentials openings
        # when a key is created accidentally without timeout (from an incr
        # operation)
        await asyncio.gather(
            self.cache.delete(self.rule.failure_cache_key),
            self.cache.delete(self.rule.request_cache_key)
        )

        self._notify_open_circuit()

    async def __aenter__(self):
        if await self.is_circuit_open():
            raise self.failure_exception

        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        await self._increase_request_count()
        if self._is_catchable_exception(exc_type):
            await self._increase_failure_count()

            total_failures, total_requests = await asyncio.gather(
                self.get_total_failures(),
                self.get_total_requests()
            )

            if self.rule.should_open_circuit(
                total_failures=total_failures,
                total_requests=total_requests
            ):
                await self.open_circuit()
                self._notify_max_failures_exceeded()
                raise self.failure_exception

    def __call__(self, func):
        @wraps(func)
        async def inner(*args, **kwargs):
            async with self:
                return await func(*args, **kwargs)
        return inner

    async def _increase_failure_count(self):
        if (
            await self.is_circuit_open() or
            not self.rule.should_increase_failure_count()
        ):
            return

        # Between the cache.add and cache.incr, the cache MAY expire,
        # which will lead to a circuit that will eventually open
        await self.cache.add(
            self.rule.failure_cache_key,
            0,
            self.failure_timeout
        )

        total_failures, total_requests = await asyncio.gather(
            self.cache.incr(self.rule.failure_cache_key),
            self.get_total_requests()
        )

        self.rule.log_increase_failures(
            total_failures=total_failures,
            total_requests=total_requests
        )

    async def _increase_request_count(self):
        if (
            await self.is_circuit_open() or
            not self.rule.should_increase_request_count()
        ):
            return

        await self.cache.add(
            self.rule.request_cache_key,
            0,
            self.failure_timeout
        )
        # To calculate the exact percentage, the cache of requests and the
        # cache of failures must expire at the same time.
        if self.rule.should_increase_failure_count():
            await self.cache.add(
                self.rule.failure_cache_key, 0, self.failure_timeout
            )

        await self.cache.incr(self.rule.request_cache_key)


circuit_breaker = CircuitBreaker



from .base import AsyncCacheAdapterBase


class Adapter(AsyncCacheAdapterBase):

    async def add(self, key, value, timeout=None):
        try:
            return await super().add(key, value, timeout)
        except ValueError:
            return True

    async def incr(self, key):
        return await self.cache.increment(key)

    async def flushdb(self):
        return await self.cache.clear()



import abc


class BaseRule(metaclass=abc.ABCMeta):

    def __init__(self, failure_cache_key, request_cache_key=None):
        self.failure_cache_key = failure_cache_key
        self.request_cache_key = request_cache_key

    @abc.abstractmethod
    def should_open_circuit(self, total_failures, total_requests):
        pass

    def should_increase_failure_count(self):
        return self.failure_cache_key is not None

    def should_increase_request_count(self):
        return self.request_cache_key is not None

    @abc.abstractmethod
    def log_increase_failures(self, total_failures, total_requests):
        pass



import logging

from .base import BaseRule

logger = logging.getLogger(__name__)


class MaxFailuresRule(BaseRule):

    def __init__(self, max_failures, failure_cache_key):
        super().__init__(failure_cache_key=failure_cache_key)
        self.max_failures = max_failures

    def should_open_circuit(self, total_failures, total_requests):
        return total_failures >= self.max_failures

    def log_increase_failures(self, total_failures, total_requests):
        logger.info(
            f'Increase failure for: {self.failures_cache_key} - '
            f'max failures {self.max_failures} - '
            f'total requests {total_requests} - '
            f'total failures {total_failures}'
        )



from functools import wraps

from lasier.adapters.caches.base import CacheAdapterBase

from .base import CircuitBreakerBase


class CircuitBreaker(CircuitBreakerBase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not isinstance(self.cache, CacheAdapterBase):
            self.cache = CacheAdapterBase(self.cache)

    def is_circuit_open(self):
        return self.cache.get(self.circuit_cache_key) == 1

    def get_total_failures(self):
        return self.cache.get(self.rule.failure_cache_key) or 0

    def get_total_requests(self):
        return self.cache.get(self.rule.request_cache_key) or 0

    def open_circuit(self):
        self.cache.set(self.circuit_cache_key, 1, self.circuit_timeout)

        # Delete the cache key to mitigate multiple sequentials openings
        # when a key is created accidentally without timeout (from an incr
        # operation)
        self.cache.delete(self.rule.failure_cache_key)
        self.cache.delete(self.rule.request_cache_key)

        self._notify_open_circuit()

    def __enter__(self):
        if self.is_circuit_open():
            raise self.failure_exception

        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self._increase_request_count()
        if self._is_catchable_exception(exc_type):
            self._increase_failure_count()

            if self.rule.should_open_circuit(
                total_failures=self.get_total_failures(),
                total_requests=self.get_total_requests()
            ):
                self.open_circuit()
                self._notify_max_failures_exceeded()
                raise self.failure_exception

    def __call__(self, func):
        @wraps(func)
        def inner(*args, **kwargs):
            with self:
                return func(*args, **kwargs)
        return inner

    def _increase_failure_count(self):
        if (
            self.is_circuit_open() or
            not self.rule.should_increase_failure_count()
        ):
            return

        # Between the cache.add and cache.incr, the cache MAY expire,
        # which will lead to a circuit that will eventually open
        self.cache.add(self.rule.failure_cache_key, 0, self.failure_timeout)
        total_failures = self.cache.incr(self.rule.failure_cache_key)

        self.rule.log_increase_failures(
            total_failures=total_failures,
            total_requests=self.get_total_requests()
        )

    def _increase_request_count(self):
        if (
            self.is_circuit_open() or
            not self.rule.should_increase_request_count()
        ):
            return

        self.cache.add(self.rule.request_cache_key, 0, self.failure_timeout)
        # To calculate the exact percentage, the cache of requests and the
        # cache of failures must expire at the same time.
        if self.rule.should_increase_failure_count():
            self.cache.add(
                self.rule.failure_cache_key, 0, self.failure_timeout
            )

        self.cache.incr(self.rule.request_cache_key)


circuit_breaker = CircuitBreaker



from setuptools import find_packages, setup


def read(fname):
    with open(fname) as f:
        return f.read()


description = 'A sync/async circuit breaker implementation'
try:
    long_description = read('README.md')
except IOError:
    long_description = description


download_url = 'https://github.com/luizalabs/lasier/tarball/master'

setup(
    name='lasier',
    version='0.0.1',
    install_requires=[],
    url='https://github.com/luizalabs/lasier',
    author='Luiza Labs',
    author_email='pypi@luizalabs.com',
    keywords='circuit breaker sync async',
    description=description,
    long_description=long_description,
    long_description_content_type='text/markdown',
    download_url=download_url,
    packages=find_packages(exclude=['tests*']),
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
    ]
)




class CacheAdapterBase:

    def __init__(self, cache):
        self.cache = cache

    def add(self, key, value, timeout=None):
        return self.cache.add(key, value, timeout)

    def set(self, key, value, timeout=None):
        return self.cache.set(key, value, timeout)

    def incr(self, key):
        return self.cache.incr(key)

    def get(self, key):
        return self._convert_to_int(self.cache.get(key))

    def delete(self, key):
        return self.cache.delete(key)

    def flushdb(self):
        return self.cache.flushdb()

    def _convert_to_int(self, value):
        if value is not None:
            return int(value)
        return None


class AsyncCacheAdapterBase(CacheAdapterBase):

    async def add(self, key, value, timeout=None):
        return await self.cache.add(key, value, timeout)

    async def set(self, key, value, timeout=None):
        return await self.cache.set(key, value, timeout)

    async def incr(self, key):
        return await self.cache.incr(key)

    async def get(self, key):
        return self._convert_to_int(await self.cache.get(key))

    async def delete(self, key):
        return await self.cache.delete(key)

    async def flushdb(self):
        return await self.cache.flushdb()
